/* scopeman/scopeman.cc
 *
 * Copyright (C) 2017 Hajime Nanjo
 *
 * Application to manupulate a Tektornix oscilloscope,
 * and to recored waveform data.
 * 
 * The license term of this code is GPLv2.
 * The author's email address is 
 * nanjo@champ.hep.sci.osaka-u.ac.jp 
 */

#include <iostream>
#include <fstream>
#include <string>
#include <cstring>
#include <cstdlib>

extern "C" {
#include "../vxi11/vxi11_user.h"
}

#include "tek_scope.h"

extern "C" {
#include <readline/readline.h>
#include <readline/history.h>
}


#define BUF_LEN 100000

#define N_CMD 35

const char* cmdname[N_CMD]={
  "HORIZONTAL:MAIN:SECDIV",
  "HORIZONTAL:TRIGGER:POSITION",
  "HORIZONTAL:DELAY:STATE",
  "HORIZONTAL:DELAY:TIME",
  "HORIZONTAL:RECORDLENGTH",
  "SELECT:CONTROL",
  "SELECT:CH1",
  "SELECT:CH2",
  "SELECT:CH3",
  "SELECT:CH4",
  "CH1:VOLTS",
  "CH1:POSITION",
  "CH2:VOLTS",
  "CH2:POSITION",
  "CH3:VOLTS",
  "CH3:POSITION",
  "CH4:VOLTS",
  "CH4:POSITION",
  "TRIGGER:A:EDGE:SOURCE",
  "TRIGGER:A:EDGE:SLOPE",
  "TRIGGER:A:EDGE:SOURCE",
  "TRIGGER:A:LEVEL",
  "TRIGGER:A:MODE",
  "ACQUIRE:MODE",
  "ACQUIRE:NUMAVG",
  "ACQUIRE:STATE",
  "ACQUIRE:STOPAFTER",
  "DATA:SOURCE",
  "DATA:ENCDG",
  "GETWFM",
  "GETIMG",
  "DISPLAY:CLOCK",
  "DISPLAY:PERSISTENCE",
  "DISPLAY:PERSISTENCE:CLEAR",
  "TIMEOUT"
};

const char* hints[N_CMD]={
  "[sec/div]",
  "0-100[%]",
  "ON/OFF",
  "[sec]",
  "500/10000",
  "CH1",
  "ON/OFF",
  "ON/OFF",
  "ON/OFF",
  "ON/OFF",
  "[V]",
  "-5 - +5",
  "[V]",
  "-5 - +5",
  "[V]",
  "-5 - +5",
  "[V]",
  "-5 - +5",
  "CH1",
  "FALL/RISE",
  "EXT/CH1/LINE",
  "[V]",
  "NORMAL/AUTO",
  "SAMPLE/AVERAGE",
  "2-512 by  2^x",
  "RUN/STOP",
  "RUNSTOP/SEQUENCE",
  "CH1",
  "SRIBINARY/ASCII",
  "fname [#ofWFMs] [chs by bit]",
  "fname [1/0]",
  "[1/0]",
  "[sec]/INFINITE/AUTO/MINIMUM",
  "",
  "[ms]"
};




char* command_generator (const char* text, int state);
char** scopeman_completion (const char* text, int start, int end);
void strip(std::string& lin);
bool shift(std::string& lin, std::string& out);
bool getwfm(VXI11_CLINK *clink,char* fname,int nwfm=1,int chs=1,unsigned long timeout=10000);
bool getwfm(VXI11_CLINK *clink,std::string line,unsigned long timeout=10000);
bool getimg(char* ip,
	    int isTDS300,
	    char* fname=const_cast<char*>("Image"),
	    bool isDraw=true);
bool getimg(char* ip,
	    int isTDS300,
	    std::string line);

bool setTimeOut(std::string line,unsigned long int &timeout);




int main(int argc, char **argv)
{

  bool isDry=false;
  //bool isDry=true;

  static char	*device_ip;
  static char	*device_name;

  if (argc != 2) {
    std::cout << "usage: "
	      << argv[0]
	      << " your.inst.ip.addr" << std::endl;
    return 1;
  }


  device_ip = argv[1];
  VXI11_CLINK *clink;
  int ret=0;
  if(!isDry) {
    char *device_name =0;
    ret=vxi11_open_device(&clink, device_ip, device_name);
  }

  if (ret != 0) {
    std::cout << "Error: could not open device "
	      << device_ip
	      << " , quitting" << std::endl;
    return 2;
  }

  int isTDS3000=tek_scope_is_TDS3000(clink);
  if(isTDS3000) {
    std::cout << "Hello,this is TDS3000 oscilosocpe." << std::endl;
  } else {
    std::cout << "Hello ,this is probably MSOXXXX oscilosocpe." << std::endl;
  }

  unsigned long	timeout=10000; /* in ms (= 10 seconds) */
  std::cout << "Defualt timeout for waveform capture is " << timeout << " ms." << std::endl;
  
  char* env = std::getenv("HOME");
  std::string histfile = std::string(env);
  histfile += "/.scopeman_history";
  int state = read_history(histfile.c_str());
  if(state!=0) {
    write_history(histfile.c_str());
  }
  rl_readline_name = "ScopeMan";
  rl_attempted_completion_function = scopeman_completion;



  char* cmd = new char[256];
  char* buf = new char[BUF_LEN];

  while(1){
    std::string line = readline ("ScopeMan: ");
    strip(line);
    if( line.size()==0 ) continue;
    
    std::sprintf(cmd,"%s",line.c_str());

    if(isDry) {
      std::cout << "Input: " << cmd << std::endl;
    }
    
    add_history (cmd);
    
    if(line=="q") break;
    
    if(std::string(line,0,4) == "help") {
      std::string sub=std::string(line,4,line.size()-4);
      strip(sub);
      bool hit=false;
      for(int k=0;k<N_CMD;k++) {
	if(sub==std::string(cmdname[k])) {
	  std::cout << sub << "\t" << hints[k] << std::endl;
	  hit=true;
	  break;
	}
      }
      if(!hit) {
	for(int k=0;k<N_CMD;k++) {
	  std::cout << cmdname[k] << "\t" << hints[k] << std::endl;
	}
	std::cout << "q" << "\t" << "Quit" << std::endl;
      }
      continue;
    }

    if(isDry) continue;
      

    
    if(std::string(line,0,6)=="GETWFM") {
      bool ok=getwfm(clink,line,timeout);
    } else if(std::string(line,0,6)=="GETIMG") {
      bool ok=getimg(device_ip,isTDS3000,line);
    } else if(std::string(line,0,7)=="TIMEOUT") {
      bool ok=setTimeOut(line,timeout);
    } else {
      if (vxi11_send(clink, cmd) < 0) break;
      if (std::strstr(cmd, "?") != 0) {
	long bytes_returned = vxi11_receive(clink, buf, BUF_LEN);
	if (bytes_returned > 0) {
	  std::cout << buf << std::endl;
	} else if (bytes_returned == -15) {
	  std::cout << "*** [ NOTHING RECEIVED ] ***" << std::endl;
	} else {
	  break;
	}
      }
    }
  }

  if(!isDry) {
    ret=vxi11_close_device(clink,device_ip);
  }

  delete [] cmd;
  delete [] buf;

  append_history(500,histfile.c_str());
  history_truncate_file(histfile.c_str(),500);

  
  
  return 0;
}




















char *
command_generator (const char* text, int state)
{
  static int list_index, len;

  std::string stext = text;
  if (!state) {
    list_index = 0;
    len = stext.size();
  }
  
  while(list_index<N_CMD) {
    std::string sname=cmdname[list_index];
    list_index++;
    std::string subs=std::string(sname,0,len);
    if(subs==stext) {
      char* oc = new char[sname.size()+1];
      std::sprintf(oc,"%s",sname.c_str());
      return oc;
    }
  }
  return 0;
}

char **
scopeman_completion (const char* text, int start, int end)
{
  char **matches =0;

  if (start == 0)
    matches = rl_completion_matches (text, command_generator);

  return matches;
}

void strip(std::string& lin)
{
  if(lin.size()>0) {
    while(std::string(lin,0,1) == " ") {
      lin.replace(0,1,"");
    }
  }
  
  if(lin.size()>0) {
    while(std::string(lin,lin.size()-1,lin.size()) == " ") {
      lin.replace(lin.size()-1,1,"");
    }
  }
}


bool shift(std::string& lin, std::string& out)
{
  std::string s1,s2;
  strip(lin);
  if(lin.size()<=0)  return false;
  int num = lin.find(" ");
  if(num<=0) {
    out = lin;
    lin = "";
    return true;
  }
  
  s1=std::string(lin.begin(),lin.begin()+num);
  strip(s1);
  s2=std::string(lin.begin()+num+1,lin.end());
  strip(s2);
  lin = s2;
  out = s1;
  return true;
}


bool getwfm(VXI11_CLINK *clink,
	    char* fname,
	    int nwfm,
	    int chs,
	    unsigned long timeout)
{
  char 		wfname[1024];
  char 		wfiname[1024];

  const char* cch[4]={"CH1","CH2","CH3","CH4"};
  int active[4]={0,0,0,0};
  active[3]=chs/8;
  chs=chs-active[3]*8;
  active[2]=chs/4;
  chs=chs-active[2]*4;
  active[1]=chs/2;
  chs=chs-active[1]*2;
  active[0]=chs;
  
  
  int npoints=10000;
  //unsigned long	timeout=10000; /* in ms (= 10 seconds) */
  int clear_sweeps=1;
  char		channel[20];
  
  if (tek_scope_init(clink) !=0 ) {
    std::cout << "GETWFM failed." << std::endl;
    return false;
  }

  char	cmd[1024];
  std::sprintf(cmd,"HOR:RECORDLENGTH %d",npoints);
  vxi11_send(clink,cmd);
  long actual_npoints =vxi11_obtain_long_value(clink, "HOR:RECORDLENGTH?");
  vxi11_send(clink, "ACQUIRE:MODE SAMPLE");
  
  
  
  long buf_size = tek_scope_set_for_capture(clink, clear_sweeps, timeout);
  char* buf=new char[buf_size];
  
  bool isWfiSaved=false;
  for(int i=0;i<nwfm;i++) {
    if(i<3) {
      std::cout << "    "
		<< int(i+1)
		<< " th" << std::endl;
    } else {
      if(nwfm>9) {
	int div=10;
	if(nwfm>100) {
	  div=100;
	}
	int unit=int(nwfm/div);
	if((i+1)%unit==0) {
	  std::cout << "    "
		    << int((i+1)/unit)*100/div
		    << " %" << std::endl;
	}
      }
    }


    if(i==1) {
      //Save wfi just after single acquisition
      for(int j=0;j<4;j++) {
	if(active[j]) {
	  std::sprintf(wfiname,"%s-ch%d.wfi",fname,j+1);
	  tek_scope_write_wfi_file(clink, wfiname,
				   const_cast<char*>(cch[j]),
				   const_cast<char*>("scopeman"),
				   1, timeout);
	}
      }
      isWfiSaved=true;
    }


    

    vxi11_send(clink, "ACQUIRE:STATE 1");
    long opc_value = vxi11_obtain_long_value_timeout(clink, "*OPC?", timeout);
    if (opc_value != 1) {
      std::cout << "OPC? request returned " << opc_value <<
	" (should be 1), maybe you need a longer timeout?" << std::endl;
      std::cout << "Not grabbing any data, returning -1" << std::endl;
      return -1;
    }
    
    for(int j=0;j<4;j++) {
      if(active[j]) {
	std::sprintf(wfname,"%s-ch%d-%03d.wf",fname,j+1,i);
	std::ofstream f_wf(wfname,std::ios::binary);

	//long bytes_returned
	//= tek_scope_get_data(clink, const_cast<char*>(cch[j]),
	//clear_sweeps, buf, buf_size, timeout);

	/////////////////////////////
	//Replaced from tek_scope_get_data
	
	int ret=0;
	std::sprintf(cmd,"DATA:SOURCE %s",const_cast<char*>(cch[j]));
	ret=vxi11_send(clink, cmd);
        if(ret < 0) {
	  std::cout << "error, could not send DATA SOURCE cmd, quitting..." << std::endl;
	  return ret;
	}
	vxi11_send(clink, "CURVE?");
        long bytes_returned=vxi11_receive_data_block(clink, buf, buf_size, timeout);
	/////////////////////////////
	
	if (bytes_returned <=0) {
	  std::cout << "Problem reading the ch "
		    << j << ", "
		    << i << " th data, skipping..." << std::endl;
	  f_wf.close();
	  continue;
	}
	f_wf.write(buf, bytes_returned);
	f_wf.close();
      }
    }
  }
  delete[] buf;

  if(!isWfiSaved) {
    for(int j=0;j<4;j++) {
      if(active[j]) {
	std::sprintf(wfiname,"%s-ch%d.wfi",fname,j+1);
	tek_scope_write_wfi_file(clink, wfiname,
				 const_cast<char*>(cch[j]),
				 const_cast<char*>("scopeman"),
				 1, timeout);
      }
    }
  }
  return true;
}


bool getwfm(VXI11_CLINK *clink,std::string line,unsigned long timeout)
{
  std::string tmpout;
  std::string out[4];
  bool ok=true;
  int cnt=0;
  while(ok) {
    ok=shift(line,tmpout);
    if(ok) {
      out[cnt]=tmpout;
      cnt++;
      if(cnt>=4) break;
    }
  }
  
  char fname[1024];
  int nwfm=1;
  int chs=1;
  if(cnt>1) {
    std::sprintf(fname,"%s",out[1].c_str());
  }
  if(cnt>2) {
    nwfm=std::atoi(out[2].c_str());
  }
  if(cnt>3) {
    chs=std::atoi(out[3].c_str());
  }
  
  if(cnt>=2 && cnt<=4) {
    ok=getwfm(clink,fname,nwfm,chs,timeout);
  } else {
    std::cout << "GETWFM error" << std::endl;
  }
  if(!ok) {
    std::cout << "GETWFM error" << std::endl;
  }
}


bool getimg(char* ip,int isTDS3000,char* fname,bool isDraw)
{
  bool ok=false;
  char cmdl[1024];
  if(isTDS3000) {
    std::sprintf(cmdl,"wget -N -t 1 http://%s/image.png -O %s.png",ip,fname);
  } else {
    std::sprintf(cmdl,"wget -N -t 1 http://%s/control/image.png -O %s.png",ip,fname);
    //std::sprintf(cmdl,"wget -N -t 1 http://%s/Image.png -O %s.png",ip,fname);
  }
  int ret=std::system(cmdl);
  ok=true;
  if(ok && isDraw) {
    std::sprintf(cmdl,"display %s.png &",fname);
    ret=std::system(cmdl);
    if(ret==0) ok=true;
  }
  return ok;
}

bool getimg(char* ip,int isTDS3000,std::string line)
{
  std::string tmpout;
  std::string out;
  std::string out2;
  bool ok=shift(line,tmpout);
  if(!ok) {
    return false;
  }
  ok=shift(line,out);
  if(ok) {
    char fname[1024];
    std::sprintf(fname,"%s",out.c_str());
    
    ok=shift(line,out2);
    if(ok) {
      int iDraw=std::atoi(out2.c_str());
      if(iDraw==1) {
	ok=getimg(ip,isTDS3000,fname,true);
      } else {
	ok=getimg(ip,isTDS3000,fname,false);
      }
    } else {
      ok=getimg(ip,isTDS3000,fname);
    }
  } else {
    ok=getimg(ip,isTDS3000);
  }
  
  return ok;
}

bool setTimeOut(std::string line,unsigned long &timeout)
{
  std::string tmpout;
  std::string out;
  bool ok=shift(line,tmpout);
  if(!ok) {
    return false;
  }
  ok=shift(line,out);
  if(ok) {
    timeout=std::atoi(out.c_str());
    std::cout << "Timeout is " << timeout << " ms."<< std::endl;
  } else {
    std::cout << "Current timeout is " << timeout << " ms." << std::endl;
  }
  return true;
}
