/* scoepman/tek_scope.cc
 *
 * Copyright (C) for the modification 2017 Hajime Nanjo
 *
 * The file, tek_user.h with the revision 1.6 from tek_1.04.tar.gz 
 * provided  by S. D.Sharples was modified on 2017/06/16.
 * This program was licensed under the term of the GPLv2 license.
 * Original revision history was included elsewhere.
 * The modification was made for the update of vxi11 library,
 * where the function interfaces were changed.
 * Only some part of the original functions were included with the modification.
 * One wrapper function for vxi11 is also included for the backward compatibility,
 * which only can be used under the c++ overload functionality.
 * The author's email address is 
 * nanjo@champ.hep.sci.osaka-u.ac.jp 
 *
 * tek_user.c
 * Copyright (C) 2006-2007 Steve D. Sharples
 *
 * User library of useful functions for talking to Tektronix instruments
 * (e.g. scopes and AFG's) using the VXI11 protocol, for Linux. You will also
 * need the vxi11.tar.gz source, currently available from:
 * http://optics.eee.nottingham.ac.uk/vxi11/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation;  version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The author's email address is steve.sharples@nottingham.ac.uk
 */


#include <iostream>
#include <cstring>
#include <cstdio>
#include <cmath>

#include "tek_scope.h"

int vxi11_send(VXI11_CLINK *clink, const char *cmd) {
  size_t len=std::strlen(cmd);
  return vxi11_send(clink,cmd,len);
}

/* Returns 1 if the scope is a TDS3000 series, 0 otherwise. Used to check on
 * the availability of the HOR:MAIN:SAMPLERATE? query */
int tek_scope_is_TDS3000(VXI11_CLINK *clink) {
  char	buf[256];
  vxi11_send_and_receive(clink, "*IDN?", buf, 256, VXI11_READ_TIMEOUT);
  if (std::strncmp("TDS 3",buf+10,5) == 0) return 1;
  //if (strncmp("MSO40",buf+10,5) == 0) {
  //	printf("it's a match!\n");
  //	return 1;
  //	}
  return 0;
}


/*****************************************************************************
 * Generic Tektronix SCOPE functions, suitable for all oscilloscopes...      *
 * ... or rather, at time of writing, suitable for all TDS3000B series and   *
 * all DPO/MSO4000 series scopes (which are the only ones we have)           *
 *****************************************************************************/

/* Set up some fundamental settings for data transfer. It's possible
 * (although not certain) that some or all of these would be reset after
 * a system reset. It's a very tiny overhead right at the beginning of your
 * acquisition that's performed just once. */
int tek_scope_init(VXI11_CLINK *clink) {
  int ret=ret=vxi11_send(clink, ":HEADER 0"); /* no headers in replies */
  if(ret < 0) {
    std::cerr << "error in tek_scope_init, could not send command ':HEADER 0" << std::endl;
    return ret;
  }
  vxi11_send(clink, ":DATA:WIDTH 2"); /* 2 bytes per data point (16 bit) */
  vxi11_send(clink, ":DATA:ENCDG SRIBINARY"); /* little endian, signed */
  return 0;
}

/* This function forces ACQ:XINC to be updated. It involves changing to RUNSTOP
 * mode, recording the current acquisition mode and no of averages, setting
 * the acquisition mode to sample temporarily, then switching back to whatever
 * mode it was in the first place. A pain in the arse, an has a small overhead
 * in terms of delay. Still, it only needs doing once, and it's better than
 * getting crap data. */
void	tek_scope_force_xincr_update(VXI11_CLINK *clink, unsigned long timeout) {
  long	value;
  int	acq_state;
  char	buf[256];
  
  /* We need to perform an acq:state 1 doing first, otherwise it could
   * just get the wrong value of hor:record. We also need to set the
   * into RUNSTOP mode (even if briefly) so that XINCR get updated too.
   * Or at least that's the theory. It turns out, that even though
   * *OPC? claims that everything's been done, there is still some kind
   * of finite delay between putting the scope into runstop mode and
   * XINCR being updated. This could affect you if your last acquisition
   * on your TDS3000 was in "Fast trigger (500 points)" mode, and you'd
   * selected on the scope "Normal (10000 points)" mode.
   * The way we get around this is by putting the scope into sample
   * mode (no averaging), then either leaving it in that mode, or
   * returning it to averaging if applicable. Seems to work ok. */
  //acq_state = tek_scope_get_averages(clink);
  //tek_scope_set_averages(clink, 0); /* set to no averaging (sample mode) */
  vxi11_send(clink, "ACQUIRE:MODE SAMPLE");
  vxi11_send(clink, "ACQUIRE:STOPAFTER RUNSTOP;:ACQUIRE:STATE 1");
  value = vxi11_obtain_long_value(clink, "*OPC?");
}

/* Makes sure that the number of points we get accurately reflects what's on
 * the screen for the given sample rate. At least that's the aim.
 * If the user has asked to "clear the sweeps", set to single sequence mode. */
long tek_scope_set_for_capture(VXI11_CLINK *clink, int clear_sweeps, unsigned long timeout) {
  long	value, no_bytes;
  int	is_TDS3000;
  
  /* There is an extra command in the DPO/MSO4000 series scoped that is
   * very useful to us. The query is "HOR:MAIN:SAMPLERATE?" and this is
   * not available on the TDS3000 series. In order to work out the number
   * of points, we need the timebase, and either the sample rate OR the
   * XINCR. Only the XINCR is available on the TDS3000 scopes.
   * Unfortunately, it seems that if you change the record length, then
   * the XINCR value only gets updated after an unspecified amount of
   * time. Since we want to give the user the option of setting this
   * before acquisition, it means that we have to prat around. In order
   * to avoid unecessary pratting around for the 4000 series scopes,
   * we ask the scope what it is and work accordingly. */
  is_TDS3000 = tek_scope_is_TDS3000(clink);
  
  if (is_TDS3000 == 1) tek_scope_force_xincr_update(clink, timeout);
  
  /* If we're not "clearing the sweeps" every time, then we need to be
   * in RUNSTOP mode... otherwise it's just going to grab the same data
   * over and over again. (If it's a TDS3000, then we've already done
   * this anyway in the pratting around waiting for XINCR to update). */
  else if (clear_sweeps == 0) {
    vxi11_send(clink, "ACQUIRE:STOPAFTER RUNSTOP;:ACQUIRE:STATE 1");
    value = vxi11_obtain_long_value(clink, "*OPC?");
  }
  
  no_bytes = tek_scope_calculate_no_of_bytes(clink, is_TDS3000, timeout);
  
  if (clear_sweeps == 1) {
    vxi11_send(clink, "ACQUIRE:STOPAFTER SEQUENCE");
  }
  
  return no_bytes;
}

/* This function, tek_scope_write_wfi_file(), saves useful (to us!)
 * information about the waveforms. This is NOT the full set of
 * "preamble" data - things like the data and other bits of info you may find
 * useful are not included, and there are extra elements like the number of
 * waveforms acquired.
 *
 * It is up to the user to ensure that the scope is in the same condition as
 * when the data was taken before running this function, otherwise the values
 * returned may not reflect those during your data capture; ie if you run
 * tek_scope_set_for_capture() before you grab the data, then either call that
 * function again with the same values, or, run this function straight after
 * you've acquired your data. */
long    tek_scope_write_wfi_file(VXI11_CLINK *clink, char *wfiname, char *captured_by, int no_of_traces, unsigned long timeout) {
  FILE	*wfi;
  double	vgain, voffset, hinterval, hoffset; /* names used in wfi file */
  double	xzero, yoff, yzero; /* names used by scope, needs translating first */
  long	data_start;
  long	no_of_bytes;
  
  int is_TDS3000 = tek_scope_is_TDS3000(clink);
  no_of_bytes = tek_scope_calculate_no_of_bytes(clink,is_TDS3000, timeout);
  
  yoff = vxi11_obtain_double_value(clink, "WFMPRE:YOFF?");
  yzero = vxi11_obtain_double_value(clink, "WFMPRE:YZERO?");
  vgain = vxi11_obtain_double_value(clink, "WFMPRE:YMULT?");
  voffset = -(yoff * vgain) + yzero;
  hinterval = vxi11_obtain_double_value(clink, "WFMPRE:XINCR?");
  data_start = vxi11_obtain_long_value(clink, "DATA:START?");
  xzero = vxi11_obtain_double_value(clink, "WFMPRE:XZERO?");
  hoffset = xzero + (((double) (data_start - 1)) * hinterval);
  wfi = fopen(wfiname,"w");
  if (wfi > 0) {
    fprintf(wfi,"%% %s\n",wfiname);
    fprintf(wfi,"%% Waveform captured using %s\n\n",captured_by);
    fprintf(wfi,"%% Number of bytes:\n%ld\n\n",no_of_bytes);
    fprintf(wfi,"%% Vertical gain:\n%g\n\n",vgain);
    fprintf(wfi,"%% Vertical offset:\n%g\n\n",-voffset);
    fprintf(wfi,"%% Horizontal interval:\n%g\n\n",hinterval);
    fprintf(wfi,"%% Horizontal offset:\n%g\n\n",hoffset);
    fprintf(wfi,"%% Number of traces:\n%d\n\n",no_of_traces);
    fprintf(wfi,"%% Number of bytes per data-point:\n%d\n\n",2); /* always 2 on Tek scopes */
    fprintf(wfi,"%% Keep all datapoints (0 or missing knocks off 1 point, legacy lecroy):\n%d\n\n",1);
    fclose(wfi);
  }
  else {
    printf("error: tek_scope_write_wfi_file: could not open %s for writing\n",wfiname);
    return -1;
  }
  
  return no_of_bytes;
}

long tek_scope_write_wfi_file(VXI11_CLINK *clink, char *wfiname, char *source, 
			      char *captured_by, int no_of_traces, unsigned long timeout) 
{
  char cmd[256];
  
  /* Check the string. If it starts with 1-4 or 'm', convert accordingly;
   * otherwise leave alone */
  tek_scope_channel_str(source);
  /* set the source channel */
  sprintf(cmd,"DATA:SOURCE %s",source);
  vxi11_send(clink, cmd);
  
  return tek_scope_write_wfi_file(clink, wfiname, captured_by, no_of_traces, timeout);
}

/* Asks the scope for the number of points in the waveform, multiplies by 2
 * (always use 2 bytes per point). This function also sets the DATA:START
 * and DATA:STOP arguments, so that, when in future a CURVE? request is sent,
 * only the data that is displayed on the scope screen is returned, rather
 * than the entire acquisition buffer. This is a PERSONAL PREFERENCE, and is
 * just the way we like to work, ie we set the timebase up so that we can see
 * the signals we are interested in; when we grab the data, we expect the same
 * amount of time that is displayed on the scope screen. Although it doesn't
 * make _acquisition_ any faster (the scope may be acquiring more points than
 * we're interested in), it does reduce bandwidth over LAN. It's mainly so
 * that we get the data we can see on the screen and nothing else, though. */
long tek_scope_calculate_no_of_bytes(VXI11_CLINK *clink, int is_TDS3000, unsigned long timeout) 
{
  char	 cmd[256];
  long	 no_acq_points;
  long	 no_points;
  double sample_rate;
  long	 start,stop;
  double xincr, hor_scale;
  
  no_acq_points = vxi11_obtain_long_value(clink, "HOR:RECORD?");
  hor_scale = vxi11_obtain_double_value(clink, "HOR:MAIN:SCALE?");
  
  if (is_TDS3000 == 1) {
    xincr = vxi11_obtain_double_value(clink, "WFMPRE:XINCR?");
    no_points = (long) round((10*hor_scale)/xincr);
  }
  else {
    sample_rate = vxi11_obtain_double_value(clink, "HOR:MAIN:SAMPLERATE?");
    no_points = (long) round(sample_rate*10*hor_scale);
  }
  
  start = ((no_acq_points - no_points)/2) + 1;
  stop  = ((no_acq_points + no_points)/2);
  /* set number of points to receive to be equal to the record length */
  sprintf(cmd, "DATA:START %ld", start);
  vxi11_send(clink, cmd);
  sprintf(cmd, "DATA:STOP %ld", stop);
  vxi11_send(clink, cmd);
  
  /*	printf("no_acq_points = %ld, xincr = %g, no_points = %ld\n",no_acq_points, xincr, no_points);
	printf("start = %ld, stop = %ld\n",start, stop);
  */
  return 2*no_points;
  
}


/*****************************************************************************
 * Utility functions. No communication with device, just useful functions    *
 *****************************************************************************/

void tek_scope_channel_str(char chan, char *source) 
{
  switch(chan) {
  case 'M' :
  case 'm' : strncpy(source,"MATH",5);
    break;
  case '1' : strncpy(source,"CH1",4);
    break;
  case '2' : strncpy(source,"CH2",4);
    break;
  case '3' : strncpy(source,"CH3",4);
    break;
  case '4' : strncpy(source,"CH4",4);
    break;
  default  : 
    break;
  }
}

void tek_scope_channel_str(char *source) 
{
  tek_scope_channel_str(source[0], source);
}
